require "util"
require("mod-gui")
require("table")
--require "style"

MOD_NAME = "TrainManager"

script.on_event(defines.events.on_player_created, on_player_created)

get_sprite_button = function(player)
	debugp("in get_sprite_button")
  local button_flow = mod_gui.get_button_flow(player)
  local button = button_flow.tm_sprite_button
  if not button then
    button = button_flow.add
    {
      type = "sprite-button",
      name = "tm_sprite_button",
      sprite = "train_manager",
      style = mod_gui.button_style
      -- tooltip = {"gui-silo-script.button-tooltip"}
    }
    button.style.visible = any
  end
  return button
end




local function on_player_created(event)
	debugp("Player Created")
	get_sprite_button(game.players[event.player_index])
end

local function on_player_joined_game(event)
	debugp("Player joined game")
	get_sprite_button(game.players[event.player_index])
end

local function onInit()
	global.trainPubSub = global.trainPubSub or {}
	global.PubSubOpened = global.PubSubOpened or {}
	if not global.trainStops then
		global.trainStops = {}
		for _,station in pairs(game.surfaces[1].find_entities_filtered({ type = 'train-stop' })) do
			local force = station.force.name
			local stopName = tostring(station.backer_name)
			global.trainStops[force] = global.trainStops[force] or {}
			global.trainStops[force][stopName] = global.trainStops[force][stopName] or {}
			table.insert(global.trainStops[force][stopName], station)
		end
	end
	debugp("In onInit")
	for _, player in pairs(game.players) do
		get_sprite_button(player)
	end
end

local function gui_open_frame(player)
	local gui = mod_gui.get_frame_flow(player)
	local frame = gui.tm_button_frame
	frame.add{
            type = "button",
            name = "subscribe_button",
			style = mod_gui.button_style,
            caption = "S"
        }
		frame.add{
            type = "button",
            name = "publish_button",
			style = mod_gui.button_style,
            caption = "P"
        }
		frame.add{
			type = "sprite-button",
			name = "priority_button",
			sprite = "priority",
			style = mod_gui.button_style
      -- tooltip = {"gui-silo-script.button-tooltip"}
		}	
		frame.add{
            type = "button",
            name = "match_button",
			style = mod_gui.button_style,
            caption = "M"
        }
		frame.add{
            type = "button",
            name = "es_button",
			style = mod_gui.button_style,
            caption = "ES"
        }
		frame.add{
            type = "button",
            name = "se_button",
			style = mod_gui.button_style,
            caption = "SE"
        }				
end

local function gui_open_subtable(player)
	local gui = mod_gui.get_frame_flow(player)
	local frame = gui.sub_table
	if not frame then return end
	frame.clear()
	local sub_title = frame.add{type = "label", caption = {"subscriptions-title"}, style = "caption_label"}
	local subscriptions = frame.add{type = "table", name = "subscriptions", column_count = 2, style = "PubSub_table_style"}	
	
	for _,subs in pairs(global.subscriptions) do
		subscriptions.add{type = "label", caption = subs.backer_name}
		subscriptions.add{type = "label", caption = subs.train.id}
	end
end

local function gui_open_pubtable(player)
	local gui = mod_gui.get_frame_flow(player)
	local frame = gui.pub_table
	if not frame then return end
	frame.clear()	
	local pub_title = frame.add{type = "label", caption = {"publish-title"}, style = "caption_label"}
	local requests = frame.add{type = "table", name = "requests", column_count = 2, style = "PubSub_table_style"}	
	if global.publishers then
		for _,req in pairs(global.publishers) do
			requests.add{type = "label", caption = req.backer_name}
		-- requests.add{type = "label", caption = subs.train.id}
		end	
	end
end

local function gui_open_estable(player)
	local gui = mod_gui.get_frame_flow(player)
	local frame = gui.es_table
	if not frame then return end
	frame.clear()
	local es_title = frame.add{type = "label", caption = "Entity/Stations", style = "caption_label"}
	local es = frame.add{type = "table", name = "es", column_count = 2, style = "PubSub_table_style"}	
	
	for _,rows in pairs(global.entitystation) do
		es.add{type = "label", caption = rows.name}
		es.add{type = "label", caption = rows.backer_name}
	end
end

local function gui_open_setable(player)
	local gui = mod_gui.get_frame_flow(player)
	local frame = gui.se_table
	if not frame then return end
	frame.clear()
	local se_title = frame.add{type = "label", caption = "Open Requests", style = "caption_label"}
	local se = frame.add{type = "table", name = "se", column_count = 1, style = "PubSub_table_style"}	
	if global.requests then
		for _,rows in pairs(global.requests) do
			se.add{type = "label", caption = tostring(rows.backer_name)}
		end
	end
end

local function stationlist(priority)
	local temp = ""
		-- temp = table.concat(priority.station)
	if priority.station then
		for i, station in ipairs(priority.station) do
			temp = temp .. station[2] .. "  :  "
		end	
	end
	return temp	
end

local function gui_open_pri_frame(player)
	local gui = mod_gui.get_frame_flow(player)
	local frame = gui.pri_frame	
	if not frame then return end
	frame.clear()
	local pri_title = frame.add{type = "label", caption = "Subscription Priorities", style = "caption_label"}
	local ptable = frame.add{type = "table", name = "ptable", column_count = 3, style = "PubSub_table_style"}	
	ptable.add{type = "label", caption = "Resource"}
	ptable.add{type = "label", caption = "Id"}
	ptable.add{type = "label", caption = "Station List"}
	global.priority = global.priority or {}
	for n, priority in ipairs(global.priority) do
		debugp("n" .. n .. "res" .. priority.resource.name .. " : id" .. priority.id.name)
		local resource = ptable.add{type = "choose-elem-button", name = "Resource" .. n, elem_type = "signal", signal = {type = priority.resource.type, name = priority.resource.name}}
		resource.locked = true
		local id = ptable.add{type = "choose-elem-button", name = "Id" .. n, elem_type = "signal", signal = {type = priority.id.type, name = priority.id.name}}
		id.locked = true
		-- local temp = ""
		-- -- temp = table.concat(priority.station)
		-- for i, station in ipairs(priority.station) do
			-- temp = temp .. station[2] .. "  :  "
			-- debugp(i .. "  :  " .. temp)
		-- end
		ptable.add{type = "label", name = "Stations" .. n, caption = stationlist(priority)}
	end
	ptable.add{type = "choose-elem-button", name = "Resource", elem_type = "signal"}
	ptable.add{type = "choose-elem-button", name = "Id", elem_type = "signal"}
	ptable.add{
        type = "button",
        name = "select_station_button",
		style = mod_gui.button_style,
        caption = "-- Select --"
        }		
end

local function gui_open_station_frame(player)
	local gui = mod_gui.get_frame_flow(player)
	local frame = gui.station_frame	

	if not frame then return end
	frame.clear()
	local station_title = frame.add{type = "label", caption = "Select Stations", style = "caption_label"}	
	-- list = {}
	-- for _, es in pairs(global.entitystation) do
		-- table.insert(list, {"",es.backer_name})
	-- end
	for _, stationX in pairs(station) do
		frame.add{type = "label", caption = stationX}
	end
	frame.add{type= "drop-down", name = "station_dd", items = list}
	frame.add{
	    type = "button",
        name = "save_station_button",
		style = mod_gui.button_style,
        caption = "Save"
        }	
end

local function on_gui_elem_changed(event)
    local element = event.element
    local player = game.players[event.player_index]	
	resource = resource or {}
	-- resource.elem_value = resource.elem_value or {}
	id = id or {}
	-- id.elem_value = id.elem_value or {}
	
	if element.name == "Resource" then
		-- resource = table.deepcopy(element)
		resource.elem_type = element.elem_type
		resource.type = element.elem_value.type
		resource.name = element.elem_value.name
		debugp("resource" .. resource.name)
	elseif element.name == "Id" then
		-- id = table.deepcopy(element)
		id.elem_type = element.elem_type
		id.type = element.elem_value.type
		id.name = element.elem_value.name
		debugp("Id".. id.name)
	elseif element.name == "Schema" then
		debugp(player.opened.name)
		-- local publisher = global.publishers[player.opened]
		for _,publisher in pairs(global.publishers) do
			if publisher.entity == player.opened then
				--debugp("found")
				publisher.priority = publisher.priority or {}
				publisher.priority.elem_type = element.elem_type
				publisher.priority.type = element.elem_value.type
				publisher.priority.name = element.elem_value.name
				gui_close(player, player.opened)
				gui_open(player, player.opened)
				break
			end
		end
	end
end

local function on_gui_selection_state_changed(event)
    local element = event.element
    local player = game.players[event.player_index]	
	local item = list[element.selected_index]
	
	if element.name == "station_dd" then
		debugp(item)
		-- local x = global.entitystation[element.selected_index].backer_name
		-- debugp(x)
		table.insert(station, item)
		-- station[#station+1] = table.deepcopy(item)
	end
	gui_open_station_frame(player)
end

local function on_gui_click(event)
    local element = event.element
    local player = game.players[event.player_index]
	local gui = mod_gui.get_frame_flow(player)
	local frame = gui.tm_button_frame

    if element.name == "tm_sprite_button" then
	    if frame then
			frame.destroy()
			return
		end
		frame = gui.add{
			type = "frame",
			name = "tm_button_frame",
			direction = "horizontal",
			-- caption = {"space-progress-frame-title"},
			style = mod_gui.frame_style
		}		
        gui_open_frame(player)

    end
	if element.name == "subscribe_button" then
		local subtable = gui.sub_table
		if subtable then
			subtable.destroy()
			return
		end
		subtable = gui.add{	
			type = "frame",
			name = "sub_table",
			direction = "vertical",
			style = mod_gui.frame_style
			}
		gui_open_subtable(player)
	end
	if element.name == "publish_button" then
		local pubtable = gui.pub_table
		if pubtable then
			pubtable.destroy()
			return
		end
		pubtable = gui.add{	
			type = "frame",
			name = "pub_table",
			direction = "vertical",
			style = mod_gui.frame_style
			}
		gui_open_pubtable(player)
	end 
	if element.name == "es_button" then
		local estable = gui.es_table
		if estable then
			estable.destroy()
			return
		end
		estable = gui.add{	
			type = "frame",
			name = "es_table",
			direction = "vertical",
			style = mod_gui.frame_style
			}
		gui_open_estable(player)
	end
	if element.name == "se_button" then
		local setable = gui.se_table
		if setable then
			setable.destroy()
			return
		end
		setable = gui.add{	
			type = "frame",
			name = "se_table",
			direction = "vertical",
			style = mod_gui.frame_style
			}
		gui_open_setable(player)
	end
	if element.name == "priority_button" then
		local pri_frame = gui.pri_frame
		if pri_frame then
			pri_frame.destroy()
			return
		end
		pri_frame = gui.add{
			type = "frame",
			name = "pri_frame",
			direction = "vertical",
			style = mod_gui.frame_style
			}
		debugp("clearing resource")
		resource = {}
		id = {}
		gui_open_pri_frame(player)
	end
	if element.name == "select_station_button" then
		local station_frame = gui.station_frame
		if station_frame then
			station_frame.destroy()
			return
		end
		station_frame = gui.add{
			type = "frame",
			name = "station_frame",
			direction = "vertical",
			style = mod_gui.frame_style
			}
		debugp("Clearing station")
		station = {}
		list = {}
		for _, es in pairs(global.entitystation) do
			table.insert(list, {"",es.backer_name})
		end
		gui_open_station_frame(player)
	end	
	if element.name == "save_station_button" then
		-- global.priority = global.priority or {}
		debugp(resource.name .. " : " .. id.name)
		for i, stationX in pairs(station) do
			debugp(i .. stationX[i])
		end
		local value = {resource = resource, id = id, station = station}
		global.priority[#global.priority+1] = table.deepcopy(value)
		debugp("global priority size" .. #global.priority)
		gui.station_frame.destroy()
		gui_open_pri_frame(player)
	end
		
end

script.on_event(defines.events.on_player_created, on_player_created )
script.on_event(defines.events.on_player_joined_game, on_player_joined_game )
script.on_init(onInit)
script.on_event(defines.events.on_gui_click, on_gui_click )
script.on_event(defines.events.on_gui_elem_changed, on_gui_elem_changed)
script.on_event(defines.events.on_gui_selection_state_changed, on_gui_selection_state_changed)
local function isPS(entity)
	if (entity.name == "train-subscriber" or entity.name == "train-publisher" or entity.name == "train-counter") then
		return true
	end
	return false
end

-- function TCKey(entity)
  -- return entity.surface.index .. ":" .. entity.position.x..":"..entity.position.y
-- end

script.on_event(defines.events.on_built_entity, function(event)
	if isPS(event.created_entity) then
		debugp("PS entity")
		addPSToTable(event.created_entity)
	end
	-- if event.created_entity.type == "train-subscriber" then
		
	if event.created_entity.type == "train-stop" then
		local force = event.created_entity.force.name
		local stopName = tostring(event.created_entity.backer_name)
		global.trainStops[force] = global.trainStops[force] or {}
		global.trainrStops[force][stopName] = global.trainStops[force][stopName] or {}
		table.insert(global.trainStops[force][stopName], event.created_entity)
	end
	if event.created_entity.name == "train-counter" then
		event.created_entity.operable = false
	end
	debugp(event.created_entity.name)
end)



script.on_event(defines.events.on_robot_built_entity, function(event)
	if isPS(event.created_entity) then
		addPSToTable(event.created_entity)
	end
	if event.created_entity.type == "train-stop" then
		local force = event.created_entity.force.name
		local stopName = tostring(event.created_entity.backer_name)
		global.trainStops[force] = global.trainStops[force] or {}
		global.trainStops[force][stopName] = global.trainStops[force][stopName] or {}
		table.insert(global.trainStops[force][stopName], event.created_entity)
	end
	if event.created_entity.name == "train-counter" then
		event.created_entity.operable = false
	end
end)

local function on_preplayer_mined_item(event)
	if isPS(event.entity) then
		removePSFromTable(event.entity)
	end
	-- if event.entity.type == "train-stop" then
		-- local force = event.entity.force.name
		-- local stopName = tostring(event.entity.backer_name)
		-- for i, trainStop in ipairs(global.trainSenderStops[force][stopName]) do
			-- if trainStop.position.x == event.entity.position.x and trainStop.position.y == event.entity.position.y then
				-- table.remove(global.trainSenderStops[force][stopName], i)
				-- break
			-- end
		-- end
	-- end
end

script.on_event(defines.events.on_pre_player_mined_item, on_preplayer_mined_item )

script.on_event(defines.events.on_robot_pre_mined, function(event)
	if isPS(event.entity) then
		removePSFromTable(event.entity)
	end
	-- if event.entity.type == "train-stop" then
		-- local force = event.entity.force.name
		-- local stopName = tostring(event.entity.backer_name)
		-- for i, trainStop in ipairs(global.trainSenderStops[force][stopName]) do
			-- if trainStop.position.x == event.entity.position.x and trainStop.position.y == event.entity.position.y then
				-- table.remove(global.trainSenderStops[force][stopName], i)
				-- break
			-- end
		-- end
	-- end
end)

script.on_event(defines.events.on_entity_died, function(event)
	if isPS(event.entity) then
		removePSFromTable(event.entity)
	end
	-- if event.entity.type == "train-stop" then
		-- local force = event.entity.force.name
		-- local stopName = tostring(event.entity.backer_name)
		-- for i, trainStop in ipairs(global.trainSenderStops[force][stopName]) do
			-- if trainStop.position.x == event.entity.position.x and trainStop.position.y == event.entity.position.y then
				-- table.remove(global.trainSenderStops[force][stopName], i)
				-- break
			-- end
		-- end
	-- end
end)

local function on_train_changed_state(event)
	if event.old_state == defines.train_state.arrive_station then
		local train = event.train
		local station = train.station
		local status,err = pcall(function()
		if global.stationentity[station.backer_name] then
			table.insert(global.subscriptions, {backer_name = station.backer_name, station = station, train = train})
		end
		end)
		if not status then
			game.players[1].print(err)
		end		
	end
end

script.on_event(defines.events.on_train_changed_state, on_train_changed_state )



-- script.on_event(defines.events.on_entity_renamed, function(event)
	-- if event.entity.type == "train-stop" then
		-- local force = event.entity.force.name
		-- local stopName = tostring(event.entity.backer_name)
		-- global.trainSenderStops[force] = global.trainSenderStops[force] or {}
		-- global.trainSenderStops[force][stopName] = global.trainSenderStops[force][stopName] or {}
		-- table.insert(global.trainSenderStops[force][stopName], event.entity)
		-- for i, trainStop in ipairs(global.trainSenderStops[force][event.old_name]) do
			-- if trainStop.position.x == event.entity.position.x and trainStop.position.y == event.entity.position.y then
				-- table.remove(global.trainSenderStops[force][event.old_name], i)
				-- break
			-- end
		-- end
	-- end
-- end)

function gui_open(player, controller)
	-- if controller.name == "train-sender" then
		-- local senderData = getTCFromTable(controller)
		-- local frame = player.gui.left.add({ type = "frame", name = "train-sender", direction = "vertical"})
		-- frame.caption = {"train-controller.train-sender"}
		-- frame.add({ type = "label", name = "label", caption = {"train-controller.train-sender-target"} })
		-- frame.add({ type = "textfield", name = "target", text = senderData.target } )
	-- end
	-- debugp("in gui open")
	-- debugp(controller.name .. "in gui_open")
	if controller.name == "train-publisher" then
		local gui = mod_gui.get_frame_flow(player)
		local frame = gui.add({ type = "frame", name = "train_publisher", direction = "vertical"})
		frame.caption = {"train-controller.train-publisher"}
		frame.add({ type = "label", name = "label", caption = {"train-controller.train-publisher-target"} })
		for _,publisher in pairs(global.publishers) do
			if publisher.entity == controller then
				if publisher.priority.name then
					frame.add{type = "choose-elem-button", name = "Schema", elem_type = "signal", signal = {type = publisher.priority.type, name = publisher.priority.name}}
					for i,priority in pairs(global.priority) do
						if publisher.priority.name == priority.id.name then
							frame.add{type = "label", name = "StationList", caption = stationlist(priority)}
							return
						end
						frame.add{type = "label", name = "StationList", caption = "not defined"}
					end
					return
				end
				break
			end
		end
		frame.add{type = "choose-elem-button", name = "Schema", elem_type = "signal"}
	end
	-- if controller.name == "train-caller" then
		-- local callerData = getTCFromTable(controller)
		-- local frame = player.gui.left.add({ type = "frame", name = "train-caller", direction = "vertical"})
		-- frame.caption = {"train-controller.train-caller"}
		-- frame.add({ type = "label", name = "label", caption = {"train-controller.train-caller-target"} })
		-- -- frame.add({ type = "table", name = "rows", colspan = 1 } )
		-- frame.add({ type ="textfield", name = "target", text = callerData.target })
		-- frame.add({ type ="textfield", name = "target1", text = callerData.target1 })		
	-- end	
end

function gui_close(player, controller)
	-- if controller.name == "train-sender" then
		-- local sendtgt = player.gui.left["train-sender"]["target"]
		-- local senderData = getTCFromTable(controller)
		-- senderData.target = sendtgt.text
		-- player.gui.left["train-sender"].destroy()
	-- end
	if controller.name == "train-publisher" then
		local gui = mod_gui.get_frame_flow(player)
		local frame = gui.train_publisher
		-- local calltgt = player.gui.left["train-caller"]["target"]
		-- local callerData = getTCFromTable(controller)
		-- callerData.target = calltgt.text
		frame.destroy()
	end
	-- if controller.name == "train-caller" then
		-- local calltgt = player.gui.left["train-caller"]
		-- local callerData = getTCFromTable(controller)
		-- callerData.target = calltgt["target"].text
		-- callerData.target1 = calltgt["target1"].text
		-- player.gui.left["train-caller"].destroy()
	-- end	
end

-- function get_trains_for(station)
	-- local key = TCKey(station)
	-- if trainCounts[station.force.name] == nil then
		-- local trains = station.force.get_trains()
		-- trainCounts[station.force.name] = {}
		-- for i = 1,#trains do
			-- if trains[i].path_end_stop ~= nil then
				-- local stopKey = TCKey(trains[i].path_end_stop)
				-- trainCounts[station.force.name][stopKey] = (trainCounts[station.force.name][stopKey] or 0) + 1
			-- end
			-- if trains[i].station ~= nil then
				-- local stopKey = TCKey(trains[i].station)
				-- trainCounts[station.force.name][stopKey] = (trainCounts[station.force.name][stopKey] or 0) + 1
			-- end
		-- end
	-- end
	-- return trainCounts[station.force.name][key] or 0
-- end

script.on_event(defines.events.on_tick, function(event)
	-- onInit()
	-- trainCounts = {}
	local current_tick = event.tick
	if current_tick%60==45  then
		-- for i, trainController in ipairs(global.trainControllers) do
			-- updateTrainController(trainController, current_tick)
		-- end
		if global.counters then
			for i, counter in pairs(global.counters) do
				updateCounters(counter)
			end
		end
	end
	if current_tick%60==49 then
		if global.publishers then
			for i, publisher in pairs(global.publishers) do
				updatePublishers(publisher)
			end
		end
	end
   
   
	if current_tick%10==9  then
		local status,err = pcall(function()
		global.controllerOpened = global.controllerOpened or {}
			for pi, player in pairs(game.players) do
				if player.connected then
					if player.opened ~= nil then					
						if not global.controllerOpened[player.name] then
							gui_open(player, player.opened)
							global.controllerOpened[player.name] = player.opened
						end
					end
					if global.controllerOpened[player.name] and player.opened == nil then
						gui_close(player, global.controllerOpened[player.name])
						global.controllerOpened[player.name] = nil
					end
				end
			end
		end)
		if not status then
			-- game.players[1].print(err)
		end
	end
end)

-- script.on_init(function()
	-- onInit()
-- end)

-- function onInit()
	-- global.trainControllers = global.trainControllers or {}
	-- global.controllerOpened = global.controllerOpened or {}
	-- if not global.trainSenderStops then
		-- global.trainSenderStops = {}
		-- for _,station in pairs(game.surfaces[1].find_entities_filtered({ type = 'train-stop' })) do
			-- local force = station.force.name
			-- local stopName = tostring(station.backer_name)
			-- global.trainSenderStops[force] = global.trainSenderStops[force] or {}
			-- global.trainSenderStops[force][stopName] = global.trainSenderStops[force][stopName] or {}
			-- table.insert(global.trainSenderStops[force][stopName], station)
		-- end
	-- end
-- end

function addPSToTable(entity)
	debugp("in addPStotable" .. entity.name)
	global.entitystation = global.entitystation or {}
	global.stationentity = global.stationentity or {}
	if entity.name == "train-subscriber" then
			debugp("in insert into train.pubsub")
		local trainPubSub = {entity = entity, target=""}
		-- table.insert(global.trainPubSub, {entity = entity, target=""})
		table.insert(global.trainPubSub, trainPubSub)
		debugp(#global.trainPubSub)
		for _,station in pairs(entity.surface.find_entities_filtered{area={{x=entity.position.x - 2,y=entity.position.y - 2},{x=entity.position.x + 2,y=entity.position.y + 2}},type='train-stop'}) do			
			global.entitystation[entity] = station
			global.stationentity[station.backer_name] = true
			for i, stationX in pairs(entity.surface.find_entities_filtered{area={{x=entity.position.x - 50,y=entity.position.y - 50},{x=entity.position.x + 50,y=entity.position.y + 50}},type='train-stop',}) do
				if stationX.backer_name == station.backer_name then 
					local trains = stationX.get_train_stop_trains()
					if trains == {} then
						table.insert(global.subscriptions, {entity = entity, backer_name = station.backer_name, station = stationX, train = {}})
					end
					for i,train in pairs(trains) do
						if train.station == stationX then
							debugp(train.id)
							global.subscriptions = global.subscriptions or {}
							-- check request table
							if not check_req(station.backer_name, train) then
								table.insert(global.subscriptions, {backer_name = station.backer_name, station = stationX, train = train})
							end
						end
					end
				end
			end
		end
	end
	if entity.name == "train-publisher" then
		table.insert(global.trainPubSub, {entity = entity, target="", target1="", called=false})
		global.publishers = global.publishers or {}
		for _,station in pairs(entity.surface.find_entities_filtered{area={{x=entity.position.x - 2,y=entity.position.y - 2},{x=entity.position.x + 2,y=entity.position.y + 2}},type='train-stop'}) do
			table.insert(global.publishers, {entity = entity, backer_name = station.backer_name, enabled = false, priority = {}, request = false})
			debugp("Publisher found at" .. station.backer_name)
		end
	end
	if entity.name == "train-counter" then
		table.insert(global.trainPubSub, {entity = entity})
		global.counters = global.counters or {}
		for _,station in pairs(entity.surface.find_entities_filtered{area={{x=entity.position.x - 2,y=entity.position.y - 2},{x=entity.position.x + 2,y=entity.position.y + 2}},type='train-stop'}) do
			table.insert(global.counters, {entity = entity, backer_name = station.backer_name, station = station})
		end		
	end
end

function removePSFromTable(entity)
	for i, trainPubSub in ipairs(global.trainPubSub) do
		if notNil(trainPubSub, "entity") then
			table.remove(global.trainPubSub, i)
			break
		end
	end
	if entity.name == "train-subscriber" then
		for i, subs in ipairs(global.subscriptions) do
			if subs.entity == entity then
				table.remove(global.subscriptions, i)
				break
			end
		end
		if notNil(global.entitystation, entity) then	
			local station = global.entitystation[entity].station.backer_name
			table.remove(global.entitystation, entity)
			table.remove(global.stationentity, station)
		-- global.entitystation[entity] = nil
		-- global.stationentity[station] = nil
		end
	end
	if entity.name == "train-publisher" then	
		for i, pubs in ipairs(global.publishers) do
			if pubs.entity == entity then
				table.remove(global.publishers, i)
				break
			end
		end	
		if global.requests then
			for i, reqs in ipairs(global.requests) do
				if reqs.entity == entity then
					table.remove(global.requests, i)
					break
				end
			end	
		end
	end
	if entity.name == "train-counter" then
		for i, counter in ipairs(global.counters) do
			if counter.entity == entity then
				table.remove(global.counters, i)
				break
			end
		end
	end
end

-- function getTCFromTable(entity)
	-- for i, trainController in ipairs(global.trainControllers) do
		-- if notNil(trainController, "entity") then
			-- if trainController.entity.position.x == entity.position.x and trainController.entity.position.y == entity.position.y then
				-- return trainController
			-- end
		-- end
	-- end
-- end

function debugp(text)
	for _, player in pairs(game.players) do
		player.print(text)
	end
end

local function check_req(backer_name, train)
	global.requests = global.requests or {}
	for i, request in ipairs(global.requests) do
		for j,priority in pairs(global.priority) do
			if request.priority.name == priority.id.name then
				for _, station in ipairs(priority.station) do
					if backer_name == station[2] then
						local schedule = {}
						schedule.records = {}
						for i = 1,train.schedule.current do
							schedule.records[#schedule.records+1] = train.schedule.records[i]
						end		
						local record = {station = backer_name, wait_conditions = {}}
						record.wait_conditions[#record.wait_conditions+1] = {type = "empty", compare_type = "and"}
						schedule.records[#schedule.records+1] = record	
						schedule.current = #schedule.records
						train.manual_mode = true
						train.schedule = schedule
						train.manual_mode = false	
						table.remove(global.requests, i)
						return true
					end
				end
			end
		end
	end
	return false
end

local function match_req(publisher)
	-- Look up subscriptions table vs priorities
	debugp("In matching")
	if not global.subscriptions then return false end
	debugp("Subscriptions exist")
	for i,priority in pairs(global.priority) do
		if publisher.priority.name == priority.id.name then
			if not priority.station then return false end
			debugp("Selected priority stations exist")
			for _, station in ipairs(priority.station) do
				debugp(station[2])
				for i, subscription in ipairs(global.subscriptions) do
					debugp(subscription.backer_name)
					if subscription.backer_name == station[2] then
						debugp("Assign" .. subscription.train.id)
						local schedule = {}
						schedule.records = {}
						for i = 1,subscription.train.schedule.current do
							schedule.records[#schedule.records+1] = subscription.train.schedule.records[i]
						end
						local record = {station = publisher.backer_name, wait_conditions = {}}
						record.wait_conditions[#record.wait_conditions+1] = {type = "empty", compare_type = "and"}
						schedule.records[#schedule.records+1] = record	
						schedule.current = #schedule.records
						subscription.train.manual_mode = true
						subscription.train.schedule = schedule
						subscription.train.manual_mode = false
						table.remove(global.subscriptions, i)
						return true
					end
				end
			end
			break
		end
	end
end

function updatePublishers(publisher)
	if publisher.entity.get_or_create_control_behavior().disabled then
		-- publisher.enabled = false
		return
	end
	if publisher.request then return end
	if not publisher.enabled then
		if not publisher.priority then return end
		if not publisher.priority.name then return end
		debugp("pub name is " .. publisher.priority.name)
		local matchreq = match_req(publisher)
		if not matchreq then
			-- append a requestor record
			global.requests = global.requests or {}
			if global.requests then
				publisher.request = true
				table.insert(global.requests, publisher)
			end
		end
		-- publisher.enabled = true
	end
end

function updateCounters(counter)	
	local cb = counter.entity.get_or_create_control_behavior()
	-- if trainController.entity.name == "train-counter" then
		-- local cb = trainController.entity.get_or_create_control_behavior()
		local localstops = counter.entity.surface.find_entities_filtered{area={{x=counter.entity.position.x - 2,y=counter.entity.position.y - 2},{x=counter.entity.position.x + 2,y=counter.entity.position.y + 2}},type='train-stop'}
		if #localstops == 0 then
			cb.parameters = nil
			return
		end
		local localstop = localstops[1]
		local trains = localstop.get_train_stop_trains()
		local traincount = 0
		for _, train in pairs(trains) do
			if train.path_end_stop == localstop then
				traincount = traincount + 1
			end
		end
		cb.parameters = {parameters = {
			{
				signal = {
					type = "virtual",
					name = "train-counter"
				},
				count = traincount,
				index = 1
			}
		}}
end

function notNil(class, var)
	value = false
	pcall (function()
		if class[var] then
			value = true
		end
	end)
	return value
end

-- function updateTrainController(trainController, tick)
	-- if trainController.entity.name == "train-sender" then
		-- if not trainController.entity.get_or_create_control_behavior().disabled then
			-- for _,train in pairs(trainController.entity.surface.find_entities_filtered{area={{x=trainController.entity.position.x - 2,y=trainController.entity.position.y - 2},{x=trainController.entity.position.x + 2,y=trainController.entity.position.y + 2}},type='locomotive'}) do
				-- if train.train.state == defines.train_state.wait_station and trainController.target ~= "" then
					-- train.train.manual_mode = true
					-- -- local schedule = train.train.schedule or {}
					-- -- if not train.train.schedule then
						-- -- schedule.records = {}
					-- -- end
					
					-- -- train.train.schedule = { current = 1, records = {{station=trainController.target}}}
				-- --	if train.train.schedule.records[2] == nil then
				-- --	else
				-- --      table.remove(train.train.schedule.records,2)
				-- --	end
					-- local schedule = {}
					-- -- if not train.train.schedule then
						-- schedule.records = {}
					-- -- end
					-- for i,record in ipairs(train.train.schedule.records) do
						-- schedule.records[#schedule.records+1] = record
						-- break
					-- end
					-- local record = {station = trainController.target, wait_conditions = {}}
					-- record.wait_conditions[#record.wait_conditions+1] = {type = "circuit", compare_type = "and"}
					-- -- train.train.schedule = { current = 1, records = {record}}
					-- -- game.print{#schedule.records}
					-- schedule.records[#schedule.records+1] = record
					-- schedule.current = #schedule.records
					-- train.train.schedule = schedule
					-- -- table.insert(schedule.records,2,record)
					-- train.train.manual_mode = false
					-- break
				-- end
			-- end
		-- end
	-- end
	
	-- if trainController.entity.name == "train-caller" then
		-- trainController.cooldown = (trainController.cooldown or 0) - 1
		-- if trainController.cooldown < 0 then
			-- trainController.cooldown = 0
		-- end
		-- if not trainController.entity.get_or_create_control_behavior().disabled and trainController.cooldown == 0 then
			-- local force = trainController.entity.force.name
			-- local stops = global.trainSenderStops[force][trainController.target] 
			-- local stops1 = global.trainSenderStops[force][trainController.target1]
			-- -- if not #stops1 == nil then
			-- local status,err = pcall(function()
			-- for i = 1, #stops1 do
				-- stops[#stops+1] = stops1[i]
			-- end
			-- end)
			-- if not status then
				-- -- game.players[1].print(err)
			-- end
			-- -- end
			-- local localstops = trainController.entity.surface.find_entities_filtered{area={{x=trainController.entity.position.x - 2,y=trainController.entity.position.y - 2},{x=trainController.entity.position.x + 2,y=trainController.entity.position.y + 2}},type='train-stop'}
			-- if #localstops == 0 or stops == nil then
				-- trainController.cooldown = 120
				-- debugp("120")
				-- return
			-- end

			-- local localstop = tostring(localstops[1].backer_name)
			-- -- for _,stop in pairs(stops) do
				-- -- debugp(stop.backer_name)
			-- -- end
			-- for _,stop in ipairs(stops) do
				-- if not stop.get_or_create_control_behavior().disabled then
					-- -- debugp(stop.position.x, stop.position.y)
					-- for _,train in pairs(stop.surface.find_entities_filtered{area={{x=stop.position.x - 2,y=stop.position.y - 2},{x=stop.position.x + 2,y=stop.position.y + 2}},type='locomotive'}) do
					    -- -- debugp(train.train.state)
						-- if train.train.state == defines.train_state.wait_station and train.train.station == stop then
						    -- -- debugp("pass train wait condition")
							-- train.train.manual_mode = true
							
							-- local schedule = {}
							-- -- if not train.train.schedule then
								-- schedule.records = {}
							-- -- end
							-- for i,record in ipairs(train.train.schedule.records) do
								-- schedule.records[#schedule.records+1] = record
								-- break
							-- end
							
							-- local record = {station = localstop, wait_conditions = {}}
							-- record.wait_conditions[#record.wait_conditions+1] = {type = "empty", compare_type = "and"}
							-- schedule.records[#schedule.records+1] = record
							-- -- local current = schedule.current or 0
							-- -- table.insert(schedule.records,current+1,record)
							-- schedule.current = #schedule.records
							-- train.train.schedule = schedule
							-- -- train.train.schedule = { current = 1, records = {record}}
							-- -- table.insert(train.train.schedule.records,2,{record})
							-- -- train.train.schedule.records[#train.train.schedule.records+1] = {record}
						-- --	if train.train.schedule.records[2] == nil then
						-- --	else
						-- --		table.remove(train.train.schedule.records,2)
						-- --	end
							-- -- local record = {station = localstop, wait_conditions = {}}
							-- -- record.wait_conditions[#record.wait_conditions+1] = {type = "inactivity", compare_type = "and", ticks = 120 }
							-- -- table.insert(train.train.schedule.records,2,record)
							-- train.train.manual_mode = false
							-- trainController.cooldown = 10
							-- return
						-- end
					-- end
				-- end
			-- end
		-- end
	-- end
-- end